"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from typing import Any
from urllib.parse import unquote

from lxml.etree import XML, XMLSyntaxError

from arelle.ModelObject import ModelObject
from arelle.ModelXbrl import ModelXbrl
from arelle.UrlUtil import scheme
from arelle.ValidateFilingText import validateGraphicHeaderType
from arelle.typing import TypeGetText
from ..Const import supportedImgTypes
from ..Util import hasSvgEventAttributes

_: TypeGetText


# check image contents against mime/file ext and for Steganography
def checkImageContents(
    modelXbrl: ModelXbrl,
    imgElt: ModelObject,
    imgType: str,
    isFile: bool,
    data: bytes,
    consolidated: bool,
) -> None:
    guidance = 'ESEF.2.5.1' if consolidated else 'ESEF.4.1.3'
    if "svg" in imgType:
        try:
            checkSVGContent(modelXbrl, imgElt, data, guidance)
        except XMLSyntaxError as err:
            try:
                checkSVGContent(modelXbrl, imgElt, unquote(data), guidance)  # Try with utf-8 decoded data as in conformance suite G4-1-3_2/TC2
            except XMLSyntaxError:
                modelXbrl.error(f"{guidance}.imageFileCannotBeLoaded",
                                _("Image SVG has XML error %(error)s"),
                                modelObject=imgElt, error=err)
        except UnicodeDecodeError as err:
            modelXbrl.error(f"{guidance}.imageFileCannotBeLoaded",
                _("Image SVG has XML error %(error)s"),
                modelObject=imgElt, error=err)
    else:
        headerType = validateGraphicHeaderType(data)
        if (("gif" not in imgType and headerType == "gif") or
            ("jpeg" not in imgType and "jpg" not in imgType and headerType == "jpg") or
            ("png" not in imgType and headerType == "png")):
            imageDoesNotMatchItsFileExtension = f"ESEF.{guidance}.imageDoesNotMatchItsFileExtension"
            incorrectMIMETypeSpecified = f"ESEF.{guidance}.incorrectMIMETypeSpecified"
            if isFile:
                code = imageDoesNotMatchItsFileExtension
                message = _("File type %(headerType)s inferred from file signature does not match the file extension %(imgType)s")
            else:
                code = incorrectMIMETypeSpecified
                message = _("File type %(headerType)s inferred from file signature does not match the data URL media subtype (MIME subtype) %(imgType)s")
            modelXbrl.error(code, message,
                modelObject=imgElt, imgType=imgType, headerType=headerType,
                messageCodes=(imageDoesNotMatchItsFileExtension, incorrectMIMETypeSpecified))
        elif not any(it in imgType for it in supportedImgTypes[isFile]):
            modelXbrl.error(f"{guidance}.imageFormatNotSupported",
                            _("Images included in the XHTML document MUST be saved in PNG, GIF, SVG or JPEG formats: %(imgType)s is not supported"),
                            modelObject=imgElt, imgType=imgType)


def checkSVGContent(
    modelXbrl: ModelXbrl,
    imgElt: ModelObject,
    data: bytes | Any | str,
    guidance: str,
) -> None:
    rootElement = True
    for elt in XML(data).iter():
        if rootElement:
            if elt.tag != "{http://www.w3.org/2000/svg}svg":
                modelXbrl.error(f"{guidance}.imageFileCannotBeLoaded",
                                _("Image SVG has root element which is not svg"),
                                modelObject=imgElt)
            rootElement = False
        # Comments, processing instructions, and maybe other special constructs don't have string tags.
        if not isinstance(elt.tag, str):
            continue
        eltTag = elt.tag.rpartition("}")[2] # strip namespace
        if eltTag in ("object", "script", "audio", "foreignObject", "iframe", "image", "use", "video"):
            href = elt.get("href","")
            if eltTag in ("object", "script") or "javascript:" in href:
                modelXbrl.error(f"{guidance}.executableCodePresent",
                                _("Inline XBRL images MUST NOT contain executable code: %(element)s"),
                                modelObject=imgElt, element=eltTag)
            elif scheme(href) in ("http", "https", "ftp"):
                modelXbrl.error(f"{guidance}.referencesPointingOutsideOfTheReportingPackagePresent",
                                _("Inline XBRL instance document [image] MUST NOT contain any reference pointing to resources outside the reporting package: %(element)s"),
                                modelObject=imgElt, element=eltTag)
        if hasSvgEventAttributes(elt):
            modelXbrl.error(f"{guidance}.executableCodePresent",
                            _("Inline XBRL images MUST NOT contain executable code: %(element)s"),
                            modelObject=imgElt, element=eltTag)



"""
See COPYRIGHT.md for copyright information.
"""

from __future__ import annotations

from collections import Counter
from collections.abc import Generator
from pathlib import Path, PurePosixPath
from typing import TYPE_CHECKING

from arelle.packages import PackageValidation
from arelle.packages.PackageType import PackageType
from arelle.packages.report import ReportPackageConst as Const
from arelle.packages.report.ReportPackage import (
    getAllReportEntries,
    getPackageJson,
    getReportPackageJsonFile,
    getReportPackageTopLevelDirectory,
)
from arelle.typing import TypeGetText
from arelle.utils.validate.Validation import Validation

if TYPE_CHECKING:
    from arelle.FileSource import FileSource

_: TypeGetText


REPORT_PACKAGE_TYPE = PackageType("Report", "rpe")

REPORT_PACKAGE_ABORTING_VALIDATIONS = (
    PackageValidation.validatePackageZipFormat,
    PackageValidation.validateZipFileSeparators,
    PackageValidation.validatePackageNotEncrypted,
    PackageValidation.validateTopLevelDirectories,
    PackageValidation.validateDuplicateEntries,
    PackageValidation.validateConflictingEntries,
    PackageValidation.validateEntries,
)

REPORT_PACKAGE_NON_ABORTING_VALIDATIONS = (PackageValidation.validateMetadataDirectory,)


class ReportPackageValidator:
    def __init__(self, filesource: FileSource) -> None:
        self._filesource = filesource
        assert isinstance(self._filesource.basefile, str)
        self._reportType = Const.ReportType.fromExtension(Path(self._filesource.basefile).suffix)
        self._stld = getReportPackageTopLevelDirectory(self._filesource)
        self._reportPackageJsonFile = getReportPackageJsonFile(self._filesource, self._stld)

    def validate(self) -> Generator[Validation, None, None]:
        if self._filesource.reportPackage is not None and self._filesource.reportPackage.reportType is None:
            yield Validation.error(
                "rpe:unsupportedFileExtension",
                _("Report package has unsupported file extension."),
            )
            return
        for validation in REPORT_PACKAGE_ABORTING_VALIDATIONS:
            if error := validation(REPORT_PACKAGE_TYPE, self._filesource):
                yield error
                return
        if (error := self._validatePackageJson()) or (error := self._validateReports()):
            yield error
        return

    def _validatePackageJson(self) -> Validation | None:
        reportPackageJson = getPackageJson(self._filesource, self._reportPackageJsonFile)
        if reportPackageJson is None:
            if self._reportPackageJsonFile in (self._filesource.dir or []):
                return Validation.error(
                    "rpe:invalidJSON",
                    _("Report package JSON file must be a valid JSON file, per RFC 8259."),
                )
            elif self._reportType in {
                Const.ReportType.NON_INLINE_XBRL_REPORT_PACKAGE,
                Const.ReportType.INLINE_XBRL_REPORT_PACKAGE,
            }:
                return Validation.error(
                    "rpe:documentTypeFileExtensionMismatch",
                    _("%(reportType)s report package requires a report package JSON file"),
                    reportType="Inline"
                    if self._reportType == Const.ReportType.INLINE_XBRL_REPORT_PACKAGE
                    else "Non-Inline",
                )
            return None
        documentInfo = reportPackageJson.get("documentInfo")
        if not isinstance(documentInfo, dict):
            return Validation.error(
                "rpe:invalidJSONStructure",
                _("Report package 'documentInfo' must resolve to a JSON object: %(documentInfo)s"),
                documentInfo=documentInfo,
            )
        documentType = documentInfo.get("documentType")
        if not isinstance(documentType, str):
            return Validation.error(
                "rpe:invalidJSONStructure",
                _("Report package type 'documentInfo.documentType' must resolve to a JSON string: %(documentType)s"),
                documentType=documentType,
            )
        if documentType not in Const.SUPPORTED_DOCUMENT_TYPES or self._stld is None:
            return Validation.error(
                "rpe:unsupportedReportPackageVersion",
                _("Report package document type '%(documentType)s' is not supported."),
                documentType=documentType,
            )
        validDocumentTypeForFileExtension = True
        if documentType == Const.INLINE_XBRL_REPORT_PACKAGE_DOCUMENT_TYPE:
            validDocumentTypeForFileExtension = self._reportType == Const.ReportType.INLINE_XBRL_REPORT_PACKAGE
        elif documentType == Const.NON_INLINE_XBRL_REPORT_PACKAGE_DOCUMENT_TYPE:
            validDocumentTypeForFileExtension = self._reportType == Const.ReportType.NON_INLINE_XBRL_REPORT_PACKAGE
        elif documentType == Const.UNCONSTRAINED_REPORT_PACKAGE_DOCUMENT_TYPE:
            validDocumentTypeForFileExtension = self._reportType == Const.ReportType.UNCONSTRAINED_REPORT_PACKAGE
        else:
            return Validation.error(
                "rpe:unsupportedReportPackageVersion",
                _("Report package document type '%(documentType)s' is not supported."),
                documentType=documentType,
            )
        if not validDocumentTypeForFileExtension:
            return Validation.error(
                "rpe:documentTypeFileExtensionMismatch",
                _("Report package document type '%(documentType)s' does not match the file extension: %(reportType)s"),
                documentType=documentType,
                reportType=self._reportType.value if self._reportType is not None else None,
            )
        return None

    def _validateReports(self) -> Validation | None:
        reportEntries = getAllReportEntries(self._filesource, self._stld)
        filesourceFiles = self._filesource.dir or []
        topLevelDir = f"{self._stld}/" if self._stld else ""
        reportsDirExist = any(entry.startswith(f"{topLevelDir}/reports") for entry in filesourceFiles)
        reportPackageJsonFileExist = f"{self._stld}/{Const.REPORT_PACKAGE_FILE}" in filesourceFiles
        if not reportsDirExist and not reportPackageJsonFileExist:
            return None
        if self._reportType is not None:
            if not any(entry.startswith(f"{self._stld}/reports") for entry in self._filesource.dir or []):
                return Validation.error(
                    "rpe:missingReportsDirectory",
                    _("Report package must contain a reports directory"),
                )
            if not reportEntries:
                return Validation.error(
                    "rpe:missingReport",
                    _("Report package must contain at least one report"),
                )
            if len(reportEntries) > 1 and not any(report.isTopLevel for report in reportEntries):
                reportEntriesBySubDir = Counter(report.dir for report in reportEntries or [] if not report.isTopLevel)
                if any(subdirCount > 1 for subdirCount in reportEntriesBySubDir.values()):
                    return Validation.error(
                        "rpe:multipleReportsInSubdirectory",
                        _("Report package must contain only one report"),
                    )
            if self._reportType == Const.ReportType.NON_INLINE_XBRL_REPORT_PACKAGE:
                if len(reportEntries) > 1:
                    return Validation.error(
                        "rpe:multipleReports",
                        _("Non-inline XBRL report package must contain only one report"),
                    )
                if any(
                    PurePosixPath(entry.primary).suffix not in Const.NON_INLINE_REPORT_FILE_EXTENSIONS
                    for entry in reportEntries
                ):
                    return Validation.error(
                        "rpe:incorrectReportType",
                        _("Non-inline XBRL report package must contain only non-inline XBRL reports"),
                    )
                for report in reportEntries or []:
                    if PurePosixPath(report.primary).suffix == Const.JSON_REPORT_FILE_EXTENSION:
                        reportContent = getPackageJson(self._filesource, report.primary)
                        if reportContent is None:
                            return Validation.error(
                                "rpe:invalidJSON",
                                _("Non-inline XBRL report package must contain only valid JSON reports"),
                            )
                        reportDocumentInfo = reportContent.get("documentInfo")
                        if not isinstance(reportDocumentInfo, dict):
                            return Validation.error(
                                "rpe:invalidJSONStructure",
                                _("Report package 'documentInfo' must resolve to a JSON object: %(documentInfo)s"),
                                documentInfo=reportDocumentInfo,
                            )
                        reportDocumentType = reportDocumentInfo.get("documentType")
                        if not isinstance(reportDocumentType, str):
                            return Validation.error(
                                "rpe:invalidJSONStructure",
                                _("Report package type 'documentInfo.documentType' must resolve to a JSON string: %(documentType)s"),
                                documentType=reportDocumentType,
                            )

            elif self._reportType == Const.ReportType.INLINE_XBRL_REPORT_PACKAGE:
                if len(reportEntries) > 1:
                    return Validation.error(
                        "rpe:multipleReports",
                        _("Inline XBRL report package must contain only one report"),
                    )
                if any(
                    PurePosixPath(entry.primary).suffix not in Const.INLINE_REPORT_FILE_EXTENSIONS
                    for entry in reportEntries
                ):
                    return Validation.error(
                        "rpe:incorrectReportType",
                        _("Inline XBRL report package must contain only inline XBRL reports"),
                    )
        return None



"""
See COPYRIGHT.md for copyright information.
"""

from typing import TYPE_CHECKING, Optional

from arelle.ModelValue import QName, AnyURI
from arelle.PythonUtil import OrderedSet
from .ModelValueMore import SQName
from .XbrlProperty import XbrlProperty
from .XbrlTypes import XbrlTaxonomyType, QNameKeyType
from .XbrlTaxonomyObject import XbrlTaxonomyObject, XbrlReferencableTaxonomyObject, XbrlTaxonomyTagObject

class XbrlLabel(XbrlTaxonomyTagObject):
    taxonomy: XbrlTaxonomyType
    relatedName: QName # (required) Defines a QName that the label is associated with.
    labelType: QName # (required) A QName representing the label type of the label. This can be a taxonomy defined label type or a standard XBRL label type defined in specification.
    language: str # (required) Defines the language of the label using a valid BCP 47 [BCP47] language code.
    value: str # (required) The text of the label.
    properties: OrderedSet[XbrlProperty] # (optional) ordered set of property objects used to specify additional properties associated with the concept using the property object. Only immutable properties as defined in the propertyType object can be added to a concept.

    @property
    def _type(self):
        return self.labelType

class XbrlLabelType(XbrlReferencableTaxonomyObject):
    taxonomy: XbrlTaxonomyType
    name: QNameKeyType # (required) The name is a QName that uniquely identifies the label type object.
    uri: Optional[AnyURI] # (optional) A uri used to identify the label type of label objects for backward compatability with XBRL 2.1 taxonomies.
    dataType: Optional[AnyURI] # (optional) Indicates the dataType of the label object value property. This allows the value of the label to be constrained if required.
    allowedObjects: OrderedSet[QName] # (optional) Defines an ordered set of object types that can use the labelType.



'''
See COPYRIGHT.md for copyright information.
'''
from arelle import ViewFile, XbrlConst, XmlUtil
from collections import defaultdict

def viewFacts(modelXbrl, outfile, lang=None, labelrole=None, cols=None):
    modelXbrl.modelManager.showStatus(_("viewing facts"))
    view = ViewFacts(modelXbrl, outfile, labelrole, lang, cols)
    view.view(modelXbrl.modelDocument)
    view.close()

COL_WIDTHS = {
    "Concept": 80, # same as label
    "Label": 80,
    "Name":  40,
    "LocalName":  40,
    "Namespace":  40,
    "contextRef": 40,
    "unitRef": 40,
    "Dec": 5,
    "Prec": 5,
    "Lang": 6,
    "Value": 40,
    "EntityScheme": 40,
    "EntityIdentifier": 40,
    "Period": 40,
    "Dimensions": 60,
    # concept properties
    "ID": 40,
    "Type": 32,
    "PeriodType": 16,
    "Balance": 16,
    "Documentation": 100
    }
class ViewFacts(ViewFile.View):
    def __init__(self, modelXbrl, outfile, labelrole, lang, cols):
        super(ViewFacts, self).__init__(modelXbrl, outfile, "Fact List", lang)
        self.labelrole = labelrole
        self.cols = cols

    def view(self, modelDocument):
        if self.cols:
            if isinstance(self.cols,str): self.cols = self.cols.replace(',',' ').split()
            unrecognizedCols = []
            for col in self.cols:
                if col not in COL_WIDTHS:
                    unrecognizedCols.append(col)
            if unrecognizedCols:
                self.modelXbrl.error("arelle:unrecognizedFactListColumn",
                                     _("Unrecognized columns: %(cols)s"),
                                     modelXbrl=self.modelXbrl, cols=','.join(unrecognizedCols))
            if "Period" in self.cols:
                i = self.cols.index("Period")
                self.cols[i:i+1] = ["Start", "End/Instant"]
        else:
            self.cols = ["Label","contextRef","unitRef","Dec","Prec","Lang","Value"]
        col0 = self.cols[0]
        if col0 not in ("Concept", "Label", "Name", "LocalName"):
            self.modelXbrl.error("arelle:firstFactListColumn",
                                 _("First column must be Concept, Label, Name or LocalName: %(col1)s"),
                                 modelXbrl=self.modelXbrl, col1=col0)
        self.isCol0Label = col0 in ("Concept", "Label")
        self.maxNumDims = 1
        self.tupleDepth(self.modelXbrl.facts, 0)
        if "Dimensions" == self.cols[-1]:
            lastColSpan = self.maxNumDims
        else:
            lastColSpan = None
        self.addRow(self.cols, asHeader=True, lastColSpan=lastColSpan)
        self.setColWidths([COL_WIDTHS.get(col, 8) for col in self.cols])
        self.viewFacts(self.modelXbrl.facts, 0)

    def tupleDepth(self, modelFacts, indentedCol):
        if indentedCol > self.treeCols: self.treeCols = indentedCol
        for modelFact in modelFacts:
            if modelFact.context is not None:
                numDims = len(modelFact.context.qnameDims) * 2
                if numDims > self.maxNumDims: self.maxNumDims = numDims
            self.tupleDepth(modelFact.modelTupleFacts, indentedCol + 1)

    def viewFacts(self, modelFacts, indent):
        for modelFact in modelFacts:
            concept = modelFact.concept
            xmlRowElementName = 'item'
            attr = {"name": str(modelFact.qname)}
            lang = ""
            if concept is not None and self.isCol0Label:
                lbl = concept.label(preferredLabel=self.labelrole, lang=self.lang, linkroleHint=XbrlConst.defaultLinkRole)
                xmlCol0skipElt = False # provide label as a row element
                if concept.baseXsdType in ("string", "normalizedString"):
                    lang = modelFact.xmlLang
            else:
                lbl = (modelFact.qname or modelFact.prefixedName) # defective inline facts may have no qname
                xmlCol0skipElt = True # name is an attribute, don't do it also as an element
            cols = [lbl]
            if concept is not None:
                if modelFact.isItem:
                    for col in self.cols[1:]:
                        if col in ("Concept", "Label"): # label or name may be 2nd to nth col if name or label is 1st col
                            cols.append( concept.label(preferredLabel=self.labelrole, lang=self.lang) )
                        elif col == "Name":
                            cols.append( modelFact.qname )
                        elif col == "LocalName":
                            cols.append( modelFact.qname.localName )
                        elif col == "Namespace":
                            cols.append( modelFact.qname.namespaceURI )
                        elif col == "contextRef":
                            cols.append( modelFact.contextID )
                        elif col == "unitRef":
                            cols.append( modelFact.unitID )
                        elif col == "Dec":
                            cols.append( modelFact.decimals )
                        elif col == "Prec":
                            cols.append( modelFact.precision )
                        elif col == "Lang":
                            cols.append( lang )
                        elif col == "Value":
                            cols.append( "(nil)" if modelFact.xsiNil == "true" else modelFact.effectiveValue )
                        elif col == "EntityScheme":
                            cols.append( modelFact.context.entityIdentifier[0] )
                        elif col == "EntityIdentifier":
                            cols.append( modelFact.context.entityIdentifier[1] )
                        elif col == "Start":
                            cols.append( XmlUtil.text(XmlUtil.child(modelFact.context.period, XbrlConst.xbrli, "startDate")) )
                        elif col == "End/Instant":
                            cols.append( XmlUtil.text(XmlUtil.child(modelFact.context.period, XbrlConst.xbrli, ("endDate","instant"))) )
                        elif col == "Dimensions":
                            for dimQname in sorted(modelFact.context.qnameDims.keys()):
                                cols.append( str(dimQname) )
                                cols.append( str(modelFact.context.dimMemberQname(dimQname)) )
                        elif col == "ID":
                            cols.append( concept.id )
                        elif col == "Type":
                            cols.append( concept.typeQname )
                        elif col == "PeriodType":
                            cols.append( concept.periodType )
                        elif col == "Balance":
                            cols.append( concept.balance )
                        elif col == "Documentation":
                            cols.append( concept.label(preferredLabel=XbrlConst.documentationLabel, fallbackToQname=False, lang=self.lang, strip=True, linkroleHint=XbrlConst.defaultLinkRole) )
                elif modelFact.isTuple:
                    xmlRowElementName = 'tuple'
            self.addRow(cols, treeIndent=indent, xmlRowElementName=xmlRowElementName, xmlRowEltAttr=attr, xmlCol0skipElt=xmlCol0skipElt)
            self.viewFacts(modelFact.modelTupleFacts, indent + 1)



"""
See COPYRIGHT.md for copyright information.
"""

from typing import TYPE_CHECKING, Optional, Set, Any

from arelle.ModelValue import QName

class QNameAt(QName):
    def __init__(self, prefix: Optional[str], namespaceURI: Optional[str], localName: str, atSuffix:str = "end") -> None:
        super(QNameAt, self).__init__(prefix, namespaceURI, localName)
        self.atSuffix: str = atSuffix # The context suffix must be either @end or @start. If an @ value is not provided then the suffix defaults to @end.

class SQName(QName):
    pass # no properties added to QName, just the localName part doesn't have to be an xml identifier



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from arelle.Cntlr import Cntlr
from arelle.ValidateXbrl import ValidateXbrl
from arelle.typing import TypeGetText
from arelle.utils.validate.ValidationPlugin import ValidationPlugin
from .PluginValidationDataExtension import PluginValidationDataExtension


_: TypeGetText

EQUITY = 'Equity'
IE_PROFIT_LOSS = 'ProfitLossBeforeTax'
IE_PROFIT_LOSS_ORDINARY = 'ProfitLossOnOrdinaryActivitiesBeforeTax'
PRINCIPAL_CURRENCY = 'PrincipalCurrencyUsedInBusinessReport'
TURNOVER_REVENUE = 'DPLTurnoverRevenue'


class ValidationPluginExtension(ValidationPlugin):
    def newPluginData(self, cntlr: Cntlr, validateXbrl: ValidateXbrl | None) -> PluginValidationDataExtension:
        return PluginValidationDataExtension(
            self.name
        )



'''
See COPYRIGHT.md for copyright information.
'''
from arelle import (ViewWinGrid, )
from arelle.UiUtil import (gridBorder, gridSpacer, gridHdr, gridCell,
                     label,
                     TOPBORDER, LEFTBORDER, RIGHTBORDER, BOTTOMBORDER, CENTERCELL)

def viewTuplesGrid(modelXbrl, tabWin, tupleObjectId, lang=None):
    modelTuple = modelXbrl.modelObject(tupleObjectId)
    if modelTuple is not None:
        modelXbrl.modelManager.showStatus(_("viewing tuples {0}").format(modelTuple.localName))
        parentFacts = modelXbrl.facts
        try: # check if possible to get parent tuple facts
            parentFacts = modelTuple.getparent().modelTupleFacts
        except:
            pass
        view = ViewTuplesGrid(modelXbrl, tabWin, modelTuple, parentFacts, lang)

        # context menu
        menu = view.contextMenu()
        menu.add_cascade(label=_("Close"), underline=0, command=view.close)
        view.menuAddLangs()
        view.view()
        view.blockSelectEvent = 1
        view.blockViewModelObject = 0
        view.viewFrame.bind("<Enter>", view.cellEnter, '+')
        view.viewFrame.bind("<Leave>", view.cellLeave, '+')
        return view
    else:
        modelXbrl.modelManager.showStatus(_("viewing tuples requires selecting the tuple to report"), clearAfter=2000)

class ViewTuplesGrid(ViewWinGrid.ViewGrid):
    def __init__(self, modelXbrl, tabWin, tupleFact, parentFacts, lang):
        super(ViewTuplesGrid, self).__init__(modelXbrl, tabWin, "Tuples", True, lang)
        self.tupleFact = tupleFact
        self.tupleConcept = tupleFact.concept
        self.parentFacts = parentFacts

    def view(self):
        # remove old widgets
        self.viewFrame.clearGrid()

        # table name
        self.zAxisRows = 0
        self.dataCols = 0
        self.colHdrRows = 0
        self.dataRows = 0
        xFilters = []

        self.analyzeColHdrs(self.parentFacts, 1)
        self.colHdrTopRow = self.zAxisRows + (2 if self.zAxisRows else 1)
        self.dataFirstRow = self.colHdrTopRow + self.colHdrRows
        self.dataFirstCol = 2

        gridHdr(self.gridTblHdr, 0, 0,
                self.tupleFact.concept.label(lang=self.lang),
                anchor="nw",
                #columnspan=(self.dataFirstCol - 1),
                rowspan=(self.dataFirstRow),
                wraplength=200)
        self.xAxis(self.dataFirstCol, self.colHdrTopRow, self.colHdrTopRow + self.colHdrRows - 1,
                   self.tupleFact.modelTupleFacts, xFilters, True, True, True)
        self.bodyCells(self.dataFirstRow, 0, self.parentFacts, xFilters, True)

        # data cells

        #self.gridView.config(scrollregion=self.gridView.bbox(constants.ALL))

    def analyzeColHdrs(self, tupleFacts, depth):
        for xAxisChildObj in tupleFacts:
            isItem = xAxisChildObj.isItem
            if isItem:
                self.dataCols += 1
            if depth > self.colHdrRows: self.colHdrRows = depth
            self.analyzeColHdrs(xAxisChildObj.modelTupleFacts, depth+1) #recurse


    def xAxis(self, leftCol, topRow, rowBelow, tupleFacts, xFilters, childrenFirst, renderNow, atTop):
        parentRow = rowBelow
        noDescendants = True
        rightCol = leftCol
        widthToSpanParent = 0
        sideBorder = not xFilters
        if atTop and sideBorder and childrenFirst:
            gridBorder(self.gridColHdr, self.dataFirstCol, 1, LEFTBORDER, rowspan=self.dataFirstRow)
        for xAxisChildObj in tupleFacts:
            isItem = xAxisChildObj.isItem
            noDescendants = False
            rightCol, row, width, leafNode = self.xAxis(leftCol, topRow + 1, rowBelow, xAxisChildObj.modelTupleFacts, xFilters, # nested items before totals
                                                        childrenFirst, childrenFirst, False)
            if row - 1 < parentRow:
                parentRow = row - 1
            #if not leafNode:
            #    rightCol -= 1
            widthToSpanParent += width
            label = xAxisChildObj.concept.label(lang=self.lang)
            if childrenFirst:
                thisCol = rightCol
                sideBorder = RIGHTBORDER
            else:
                thisCol = leftCol
                sideBorder = LEFTBORDER
            if renderNow:
                columnspan = (rightCol - leftCol + (1 if isItem else 0))
                gridBorder(self.gridColHdr, leftCol, topRow, TOPBORDER, columnspan=columnspan)
                gridBorder(self.gridColHdr, leftCol, topRow,
                           sideBorder, columnspan=columnspan,
                           rowspan=(rowBelow - topRow + 1) )
                gridHdr(self.gridColHdr, leftCol, topRow,
                        label if label else "         ",
                        anchor="center",
                        columnspan=(rightCol - leftCol + (1 if isItem else 0)),
                        rowspan=(row - topRow + 1) if leafNode else 1,
                        wraplength=width,
                        objectId=xAxisChildObj.objectId(),
                        onClick=self.onClick)
                if isItem:
                    gridBorder(self.gridColHdr, thisCol, self.dataFirstRow - 1, BOTTOMBORDER)
                    xFilters.append(xAxisChildObj.concept,
                                     )
            if isItem:
                rightCol += 1
            if renderNow and not childrenFirst:
                self.xAxis(leftCol + (1 if isItem else 0), topRow + 1, rowBelow, xAxisChildObj.modelTupleFacts, xFilters, childrenFirst, True, False) # render on this pass
            leftCol = rightCol
        if atTop and sideBorder and not childrenFirst:
            gridBorder(self.gridColHdr, rightCol - 1, 1, RIGHTBORDER, rowspan=self.dataFirstRow)
        return (rightCol, parentRow, widthToSpanParent, noDescendants)

    def tupleDescendant(self, tupleParent, descendantConcept):
        for tupleChild in tupleParent.modelTupleFacts:
            if tupleChild.concept == descendantConcept:
                return tupleChild
            tupleDescendant = self.tupleDescendant(tupleChild, descendantConcept)
            if tupleDescendant is not None:
                return tupleDescendant
        return None

    def bodyCells(self, row, indent, tupleFacts, xFilters, zFilters):
        for modelTupleFact in tupleFacts:
            if modelTupleFact.concept == self.tupleConcept:
                for i, xColConcept in enumerate(xFilters):
                    fact = self.tupleDescendant(modelTupleFact, xColConcept)
                    if fact is not None:
                        value = fact.effectiveValue
                        objectId = fact.objectId()
                        justify = "right" if fact.isNumeric else "left"
                    else:
                        value = None
                        objectId = None
                    if value is not None:
                        gridCell(self.gridBody, self.dataFirstCol + i, row, value, justify=justify, width=12,
                                 objectId=objectId, onClick=self.onClick)
                    else:
                        gridSpacer(self.gridBody, self.dataFirstCol + i, row, CENTERCELL)
                    gridSpacer(self.gridBody, self.dataFirstCol + i, row, RIGHTBORDER)
                    gridSpacer(self.gridBody, self.dataFirstCol + i, row, BOTTOMBORDER)
                row += 1
        return row

    def onClick(self, event):
        self.modelXbrl.viewModelObject(event.widget.objectId)

    def cellEnter(self, *args):
        self.blockSelectEvent = 0

    def cellLeave(self, *args):
        self.blockSelectEvent = 1

    def cellSelect(self, *args):
        if self.blockSelectEvent == 0 and self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            #self.modelXbrl.viewModelObject(self.nodeToObjectId[self.treeView.selection()[0]])
            #self.modelXbrl.viewModelObject(self.treeView.selection()[0])
            self.blockViewModelObject -= 1

    def viewModelObject(self, modelObject):
        if self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            '''
            try:
                if isinstance(modelObject, ModelObject.ModelRelationship):
                    conceptId = modelObject.toModelObject.objectId()
                elif isinstance(modelObject, ModelObject.ModelFact):
                    conceptId = self.modelXbrl.qnameConcepts[modelObject.qname].objectId()
                else:
                    conceptId = modelObject.objectId()
                #node = self.objectIdToNode[conceptId]
                node = conceptId
                if self.treeView.exists(node):
                    self.treeView.see(node)
                    self.treeView.selection_set(node)
            except KeyError:
                    self.treeView.selection_set(())
            '''
            self.blockViewModelObject -= 1
